#*************************************************************************************************************
# Name: BF2:Sandbox
# Author: Elxx (http://elxx.net)
# Version 0.3a
#*************************************************************************************************************
# Conditions and Limitations: You may use this file locally, on listening and dedicated servers.
# This software is provided "as is". In no event shall the author, authors or contributors be held
# liable for  any direct, indirect, incidental, special, exemplary, or consequential damages (including)
# but not limited to, procurement of substitute good or services; loss of use, data or profits; or
# business interruption) however caused and on any theory of liability, whether in contract, strict liability
# arising in any way out of the use of this software, even if advised of the possibility of such damage.
# This file may be modified as long as this header remains intact. The BF2:Sandbox team offers no support
# for modified versions of this file.
#*************************************************************************************************************
#
# Configure the users who can use the Admin Tools in Multiplayer.
MP_LS_USERS = ['SampleUser1','Sir SampleAdmin2']
#
#*************************************************************************************************************

import host
import bf2
import bf2.Timer
import bf2.PlayerManager
import math
import sys
import re
import os
import inspect
from bf2.stats.constants import *

setattr(bf2.PlayerManager.Player, 'tracerDistance', 3)
setattr(bf2.PlayerManager.Player, 'objects', [])
setattr(bf2.PlayerManager.Player, 'objectTypes', [])
setattr(bf2.PlayerManager.Player, 'grabbed', [])
setattr(bf2.PlayerManager.Player, 'lastRotation', 0)
setattr(bf2.PlayerManager.Player, 'lastSpawned', [])
setattr(bf2.PlayerManager.Player, 'lastTemplate', "")
setattr(bf2.PlayerManager.Player, 'tracerOn', 1)
setattr(bf2.PlayerManager.Player, 'tracerDistance', 8)
setattr(bf2.PlayerManager.Player, 'gridSnap', 3)
setattr(bf2.PlayerManager.Player, 'rotSnap', 5)
setattr(bf2.PlayerManager.Player, 'autoGrab', 1)
setattr(bf2.PlayerManager.Player, 'numSpawned', [0,0])
setattr(bf2.PlayerManager.Player, 'preGrab', [])
setattr(bf2.PlayerManager.Player, 'preGrabMode', 0)
setattr(bf2.PlayerManager.Player, 'lastAction', [])
setattr(bf2.PlayerManager.Player, 'spawnDelay', 0)

# Actions
# 1 = grab [object, original position, original rotation]
# 2 = drop [object, frozen]
# 3 = delete [object, original position]
# 4 =

def ExceptionOutput():
	sys.stderr.write("\n" + "Exception Occured: " + str(sys.exc_info()[0]) + "\n")
	sys.stderr.write("Value: " + str(sys.exc_info()[1]) + "\n")
	sys.stderr.write("Line:" + str(readline(inspect.getfile(sys.exc_info()[2]),
							sys.exc_info()[2].tb_lineno)) + "\n")
	sys.stderr.write("Line #: " + str(sys.exc_info()[2].tb_lineno) + "\n")
	sys.stderr.write("File: " + str(inspect.getfile(sys.exc_info()[2])) + "\n" + "\n")
	#host.rcon_invoke("game.sayAll \"Exception Caught: " + str(sys.exc_info()[1]) + "\"")
	#host.rcon_invoke("game.sayAll \"Line " + str(sys.exc_info()[2].tb_lineno) + "\"")

def readline(filename, lineno):
	filen = re.sub('\\\\', '/', filename)
	file = open(filen, 'rU')
	lines = file.readlines()
	file.close()
	linen = lineno - 1
	line = re.sub('\s+', ' ', lines[linen])
	return line

SANDBOX_VERSION = "0.3"

class sandboxCore:
	def __init__(self):
		self.SANDBOX_VERSION = SANDBOX_VERSION
		# Set up event handlers
		host.registerGameStatusHandler(self.onGameStatusChanged)
		host.registerHandler('PlayerDeath', self.onPlayerDeathCQ)
		host.registerHandler('PlayerKilled', self.onPlayerKilledCQ)
		host.registerHandler('PlayerSpawn', self.onPlayerSpawn)
		host.registerHandler('EnterVehicle', self.onEnterVehicle)
		host.registerHandler('ExitVehicle', self.onExitVehicle)
		host.registerHandler('PlayerDisconnect', self.onPlayerDisconnect)
		host.registerHandler('ChatMessage', self.onChatMessage, 1)
		host.registerHandler('RemoteCommand', self.onRemoteCommand, 1)
		self.SPAWN_LIMIT = 200
		self.RAGDOLL_LIMIT = 15

	def initSbx(self):
		# Set up class variables
		self.object_types = []
		self.deleted_objects = []
		self.obj_freeze = {}
		self.obj_locked = {}
		self.is_spGame = 0
		self.ragdoll_objects = ['pallet_mobile','barrel_yellow_mobile','woodencrate_destructible_mobile','rock2m','ball']
		self.annoying_objects = ['convbelt','barrel_yellow_mobile','barrel_yellow']
		self.mp_objects = ['table01','barrel_green','fishingbox','barrel_yellow','pallet','woodencrate_destructible','concrete_pillar_beam','beam_01','container_cargo_red','container_cl_green','mi_sandbags_4m','construct_metal_plate','mi_plane_stair','pedestrian_railing_12m_ends','trestle01_dest','ch_stnwalls_10m_up', 'concreteblock_airstrip_endstr', 'convbelt', 'mi_sandbags_circle_half', 'woodsteel_bridge_segment', 'basket_hoops', 'concretebarrier', 'pedestrian_railing_6m_ends', 'wallhigh02', 'fence_dest', 'wirefence_6m', 'brickwall', 'fence_corrugated_3x12m', 'barrier_blocks_high_8m', 'gate_boom', 'rock2m', 'cafechair', 'wpipe_10m', 'tvset', 'chsalestable_02', 'wheelbarrow', 'trashcan_small', 'phonebooth', 'trafficlight_dest', 'trash_dumpster', 'power_station_pt_2', 'wpipe_support', 'cablepole','ch_wall_high_6m_door']
		self.nonmp_objects = ['ladder_10m_noholders','ladder_40m','rock2m']
		self.allowed_objects = ['cone_dest','sign_streetcrossing_dest','barrel_blue','pallet_mobile','barrel_yellow_mobile','woodencrate_destructible_mobile','barrel_green','fishingbox','ball','barrel_yellow','pallet','woodencrate_destructible','concrete_pillar_beam','beam_01','container_cargo_red','container_cl_green','mi_sandbags_4m','construct_metal_plate','mi_plane_stair','pedestrian_railing_12m_ends','trestle01_dest','ch_stnwalls_10m_up', 'ladder_10m_noholders', 'ladder_40m', 'concreteblock_airstrip_endstr', 'convbelt', 'mi_sandbags_circle_half', 'woodsteel_bridge_segment', 'basket_hoops', 'concretebarrier', 'pedestrian_railing_6m_ends', 'wallhigh02', 'fence_dest', 'wirefence_6m', 'brickwall', 'fence_corrugated_3x12m', 'barrier_blocks_high_8m', 'gate_boom', 'rock2m', 'cafechair', 'wpipe_10m', 'tvset', 'chsalestable_02', 'table01', 'wheelbarrow', 'trashcan_small', 'phonebooth', 'trafficlight_dest', 'trash_dumpster', 'power_station_pt_2', 'wpipe_support', 'cablepole','ch_wall_high_6m_door','supply_crate']
		self.allowed_vehicles = ['jeep_faav','ahe_ah1z','boat_rib','air_f35b','uav_pred','usthe_uh60','usaas_stinger','ats_tow']
		self.sbxTimer = None
		self.SPAWN_LIMIT = 2000
		self.RAGDOLL_LIMIT = 15
		self.p_objects = {}
		self.MP_LS_USERS = MP_LS_USERS
		self.allow_lock = True
		self.allow_grab = True
		self.allow_create = True
		self.gravity = -15
		self.timerLoaderActive = 0
		self.timerLoader = [1,0,""]
		self.timerLoaderObjects = []
		self.timerLoaderTempFreeze = {}
		self.spawnerTemplates = []
		self.lastObject = []
		self.allowCreateOverride = True
		self.sBoundaries = True
		# Set up the sandbox timer
		self.timerCounter = 0
		if not self.sbxTimer:
			self.sbxTimer = bf2.Timer(self.run_sbxTimer, 0 , 1)
			self.sbxTimer.setRecurring(0.01)
		# Debug mode?
		self.enableDebug = 0
		self.dbg("Sandbox has been initialized.")

	def dbg(self,msg):
		if self.enableDebug:
			print msg

	def onGameStatusChanged(self, status):
		if status == bf2.GameStatus.Playing:
			self.p_objects = {}
			self.object_types = []
			self.deleted_objects = []
			self.obj_freeze = {}
			self.obj_locked = {}
			self.gravity = -15
			self.timerLoaderActive = 0
			self.timerLoader = [1,0,""]
			self.timerLoaderObjects = []
			self.timerLoaderTempFreeze = {}
			self.spawnerTemplates = []
			host.rcon_invoke("physics.gravity " + str(self.gravity))
			if self.sBoundaries:
				self.enableBoundaries()
			else:
				self.disableBoundaries()
			self.dbg("Game status cleared and Sandbox reloaded.")

	def onRemoteCommand(self, playerId, cmd):
		from math import pi

		if playerId == -1: #fix for local servers
			playerId = 255
		self.dbg("Command received: [" + str(cmd) + "] from [" + str(playerId) + "]")
		p = bf2.PlayerManager.Player(playerId)
		p = self.p_objects[p.getName()]
		pPos = p.getDefaultVehicle().getPosition()
		pRot = p.getDefaultVehicle().getRotation()
		origCmd = cmd
		spacepos = origCmd.find(' ')
		if spacepos == -1: spacepos=len(cmd)
		cmd = origCmd[0:spacepos]
		cmdVar = origCmd[spacepos + 1:]
		if cmd == "btnSpawn":
			if not self.allow_create and (not self.MP_LS_USERS.count(p.getName()) and not self.allowCreateOverride):
				return 0
			if cmdVar and (not self.allowed_objects.count(cmdVar.lower()) and not self.allowed_vehicles.count(cmdVar.lower())):
				return 0
			if p.spawnDelay > 0:
				return 0
			if self.nonmp_objects.count(cmdVar.lower()) and not self.is_spGame:
				return 0
			if self.mp_objects.count(cmdVar.lower()) and not self.is_spGame:
			#if self.mp_objects.count(cmdVar.lower()):
				cmdVar += "_mp"
			pCam = self.getSoldierCam(p)
			pCamRot = pCam.getRotation()
			sPos = self.getVectoredPosition(pPos,p.tracerDistance,pRot,pCamRot)
			sRot = pRot
			self.dbg("new position: " + str(sPos[0]) + "," + str(sPos[1]) + "," + str(sPos[2]))
			self.dbg("new rotation: " + str(sRot[0]) + "," + str(sRot[1]) + "," + str(sRot[2]))
			if cmdVar:
				p.lastTemplate = str(cmdVar)
			else:
				if p.lastRotation:
					sRot = p.lastRotation
			cmdVar = p.lastTemplate
			# Create the spawner using the fancy new method
			#o = self.createObject(str(cmdVar), sPos, sRot, p.getName())
			try:
				self.createSpawnerObject(str(cmdVar), sPos, sRot, p.getName())
			except:
				ExceptionOutput()
			if p.objectTypes.count(str(cmdVar)) < 1:
				p.objectTypes.append(str(cmdVar))
			if self.object_types.count(str(cmdVar)) < 1:
				self.object_types.append(str(cmdVar))
			p.lastRotation = sRot
			if not self.is_spGame:
				if self.annoying_objects.count(str(cmdVar)):
					p.spawnDelay = 120
				else:
					p.spawnDelay = 80
			self.dbg("Created spawner!")

		if cmd == "btnGrabAlt":
			if self.is_spGame:
				if p.preGrabMode == 0:
					p.preGrabMode = 1
				else:
					o = p.preGrab[0]
					if hasattr(o,"oId"):
						host.rcon_invoke("Object.active id" + str(o.oId))
						host.rcon_invoke("Object.isVisible 1")
					p.preGrab[0].setPosition(p.preGrab[3])
					p.preGrabMode = 0
					p.preGrab = []

		if cmd == "btnGrab":
			if p.spawnDelay > 40:
				return 0
			if not self.allow_grab and not self.MP_LS_USERS.count(p.getName()):
				return 0
			if p.preGrabMode == 1 and self.is_spGame:
				o = p.preGrab[0]
				if hasattr(o,"oId"):
					host.rcon_invoke("Object.active id" + str(o.oId))
					host.rcon_invoke("Object.isVisible 1")
				p.preGrab[0].setPosition(p.preGrab[3])
				p.preGrabMode = 0
				p.preGrab = []
			shortest = [0,p.tracerDistance + 20.0]
			pCam = self.getSoldierCam(p)
			pCamRot = pCam.getRotation()
			sPos = self.getVectoredPosition(pPos,p.tracerDistance,pRot,pCamRot)
			sRot = pRot
			for v in self.object_types:
				for q in bf2.objectManager.getObjectsOfTemplate(v):
					if q.isValid():
						qDist = self.getVectorDistance(sPos,q.getPosition())
						if qDist < shortest[1]:
							shortest = [q,qDist]
			if shortest[0]:
				if not self.is_spGame and self.obj_locked.has_key(shortest[0]):
					if self.obj_locked[shortest[0]] != p.index: return 0
				self.dbg("Found closest object")
				distance = self.getVectorDistance(pPos,shortest[0].getPosition())
				frozen = 0
				if self.obj_freeze.has_key(shortest[0]):
					frozen = 1
					self.obj_freeze.pop(shortest[0])
				p.grabbed = [shortest[0],distance,1,shortest[0].getPosition(),shortest[0].getRotation()]
				if self.is_spGame:
					p.lastAction.append([1,shortest[0],shortest[0].getPosition(),shortest[0].getRotation(),frozen])
					# 1 = grab [object, original position, original rotation, original frozen]
				self.dbg("Grabbed closest object")
				p.spawnDelay = 80

		if cmd == "btnRotate":
			v = cmdVar.split(" ")
			if p.grabbed:
				if v[1] == "1": rFactor = p.rotSnap
				else: rFactor = -p.rotSnap
				obj = p.grabbed[0]
				if obj and obj.isValid():
					objRot = obj.getRotation()
					if v[0] == "A":
						newRot = (objRot[0] + rFactor, objRot[1], objRot[2])
					if v[0] == "P":
						newRot = (objRot[0], objRot[1] + rFactor, objRot[2])
					if v[0] == "R":
						newRot = (objRot[0], objRot[1], objRot[2] + rFactor)
					p.grabbed[4] = newRot
					obj.setRotation(newRot)
					p.lastRotation = newRot

		if cmd == "btnDrop":
			p.lastAction.append([2,p.grabbed[0],p.grabbed[1],0])
			# 2 = drop [object, distance]
			p.grabbed = []

		if cmd == "btnFreeze":
			if p.grabbed:
				if p.grabbed[0].isValid():
					self.obj_freeze[p.grabbed[0]] = [p.grabbed[0].getPosition(),p.grabbed[0].getRotation()]
					p.lastAction.append([2,p.grabbed[0],p.grabbed[1],1])
					p.grabbed = []

		if cmd == "btnLock":
			if p.grabbed and self.allow_lock:
				if p.grabbed[0].isValid():
					self.obj_locked[p.grabbed[0]] = p.index

		if cmd == "btnUnLock":
			if p.grabbed:
				if p.grabbed[0].isValid():
					try:
						del self.obj_locked[p.grabbed[0]]
					except:
						pass #do nothing

		if cmd == "btnDelete":
			if p.grabbed:
				if p.grabbed[0].isValid():
					o = p.grabbed[0]
					p.grabbed = []
					self.deleteObject(o)

		if cmd == "btnPush":
			if cmdVar == "1":
				p.grabbed[1] += 0.5
			else:
				p.grabbed[1] += 1.0

		if cmd == "btnPull":
			if cmdVar == "1":
				p.grabbed[1] -= 0.5
			else:
				p.grabbed[1] -= 1.0

		if cmd == "btnUndo":
			if p.lastAction and self.is_spGame:
				a = p.lastAction.pop()
				if a[0] == 1: #grab [object, original position, original rotation, original frozen]
					self.dbg("Grab undoing")
					o = a[1]
					if p.grabbed[0] == o:
						p.grabbed = []
					o.setPosition(a[2])
					o.setRotation(a[3])
					if a[4] or (not self.is_spGame):
						self.obj_freeze[o] = [a[2],a[3]]
						self.dbg("Refrozen")
				if a[0] == 2: # 2 = drop [object, distance]
					self.dbg("Drop undoing")
					o = a[1]
					p.grabbed = []
					p.grabbed = [o,a[2],1,o.getPosition(),o.getRotation()]
				if a[0] == 3: # 3 = delete [object name, original position, original rotation]
					self.dbg("Delete undoing")

					#o = a[1]
					#o.setPosition(a[2])
					#self.deleted_objects.remove(o)
					#p.numSpawned[0] += 1
					#self.onRemoteCommand(playerId,"btnUndo") #Undo once more so we re-grab the deleted object
				self.dbg("Undone.")

		if cmd == "btnTracerToggle":
			children = p.getPrimaryWeapon().getChildren()
			for child in children:
				if child.templateName == "glowshit":
					pos = child.getPosition()
					if p.tracerOn:
						child.setPosition((pos[0],pos[1] - 100.0,pos[2]))
						p.tracerOn = 0
					else:
						child.setPosition((pos[0],pos[1] + 100.0,pos[2]))
						p.tracerOn = 1

		if cmd == "btnAutoGrabToggle":
			if p.autoGrab:
				p.autoGrab = 0
				self.sendMsg(p,"Autograb disabled.")
			else:
				p.autoGrab = 1
				self.sendMsg(p,"Autograb enabled.")

		if cmd == "btnTracerDist":
			if cmdVar == "push":
				p.tracerDistance += 1
			else:
				p.tracerDistance -= 1
			children = p.getPrimaryWeapon().getChildren()
			for child in children:
				if child.templateName == "glowshit":
					pPos = p.getPrimaryWeapon().getPosition()
					pCam = self.getSoldierCam(p)
					pCamRot = pCam.getRotation()
					sPos = self.getVectoredPosition(pPos,p.tracerDistance,pRot,pCamRot)
					child.setPosition(sPos)
					self.sendMsg(p, "Tracer distance set to " + str(p.tracerDistance))

		if cmd == "btnRotSnap":
			if cmdVar == "inc":
				if p.rotSnap > 4:
					p.rotSnap += 5
				else:
					p.rotSnap += 1
			else:
				if p.rotSnap > 5:
					p.rotSnap -= 5
				elif p.rotSnap <= 5 and p.rotSnap > 1:
					p.rotSnap -= 1
			self.sendMsg(p, "Rotation factor set to " + str(p.rotSnap) + "deg")

		if cmd == "btnSnap":
			if cmdVar == "inc":
				if p.gridSnap > 0:
					p.gridSnap -= 1
				else:
					self.sendMsg(p, "Grid snap is already at maximum level.")
			else:
				if p.gridSnap < 3:
					p.gridSnap += 1
			if p.gridSnap < 3:
				self.sendMsg(p, "Grid snap set to " + str(p.gridSnap))
			else:
				self.sendMsg(p, "Grid snap disabled")

		if self.MP_LS_USERS.count(p.getName()) or self.is_spGame == 1:

			if cmd == "btnGravity" and cmdVar:
				try:
					gravFactor = int(cmdVar)
					self.gravity += gravFactor
					self.sendMsg(0,"Gravity set to " + str(self.gravity))
					host.rcon_invoke("physics.gravity " + str(self.gravity))
				except:
					ExceptionOutput()

			if cmd == "btnAdGrab" and cmdVar:
				if cmdVar == "1":
					self.allow_grab = True
					self.sendMsg(0,"Object grab enabled")
				else:
					self.allow_grab = False
					self.sendMsg(0,"Object grab disabled")

			if cmd == "btnAdLock" and cmdVar:
				if cmdVar == "1":
					self.allow_lock = True
					self.sendMsg(0,"Object lock enabled")
				else:
					self.allow_lock = False
					self.sendMsg(0,"Object lock disabled")

			if cmd == "btnAdCreate" and cmdVar:
				if cmdVar == "1":
					self.allow_create = True
					self.sendMsg(0,"Object create enabled")
				else:
					self.allow_create = False
					self.sendMsg(0,"Object create disabled")

			if cmd == "delAll" and cmdVar:
				if self.p_objects.has_key(cmdVar):
					pl = p_objects[cmdVar]
					pl.grabbed = []
					for o in pl.objects:
						self.deleteObject(o)
					sendMsg(pl,"All objects deleted")
			elif cmd == "delAll":
				p.grabbed = []
				for o in p.objects:
					self.deleteObject(o)

		if cmd == "btnNone":
			self.dbg("btnNone")

	def run_sbxTimer(self, data):
		try:
			self.timerCounter += 1
			if self.timerCounter >= 5:
				self.timerCounter = 0
			for p in bf2.playerManager.getPlayers():
				if not self.p_objects.has_key(p.getName()):
					self.p_objects[p.getName()] = p
					#self.dbg(str(p.getName()))
				#else:
				#	pl = self.p_objects[p.getName()]
				#	if not pl.isValid():
				#		self.p_objects[p.getName()] = p
			#for x in self.p_objects:
			if self.lastObject:
				oName = self.lastObject[0]
				user = self.lastObject[1]
				oPos = self.lastObject[2]
				oRot = self.lastObject[3]
				if (type(oPos) == type(('tuple','type'))) and (type(oRot) == type(('tuple','type'))):
					o = None
					for q in bf2.objectManager.getObjectsOfTemplate(str(oName)):
						if q.isValid() and user:
							p = self.p_objects[user]
							if p.isAlive():
								qDist = self.getVectorDistance(p.getDefaultVehicle().getPosition(),q.getPosition())
								if qDist <= (p.tracerDistance + 5.0):
									o = q
					if o:
						objs = host.rcon_invoke("object.listObjectsofTemplate " + str(oName))
						print objs
						objs = objs.split()
						if self.is_spGame:
							o.oId = objs[-7]
						#o.setPosition(oPos)
						#o.setRotation(oRot)
						if user:
							p = self.p_objects[user]
							p.objects.append(o)
							if p.autoGrab and (not self.allowed_vehicles.count(oName)):
								p.grabbed = [o,p.tracerDistance,1,o.getPosition(),o.getRotation()]
						self.lastObject = []
			for x in bf2.playerManager.getPlayers():
				z = bf2.PlayerManager.Player(x.index)
				#self.dbg(str(z.getName()))
				p = self.p_objects[z.getName()]
				try:
					p.getName()
					z.getName()
				except:
					try:
						self.p_objects.pop(x.getName())
					except:
						pass
					try:
						self.p_objects.pop(z.getName())
					except:
						pass
					return 0
				if p.getName() != z.getName():
					#self.sendMsg("DEBUG: Fixing player object mismatch for " + str(z.getName()))
					self.fixPlayerObject(z.index,z.getName())
					p = self.p_objects[z.getName()]
				#p = self.p_objects[x]
				if p.spawnDelay > 0:
					p.spawnDelay -= 1
				if p.isAlive():
					#p = self.p_objects[p.index]
					pPos = p.getDefaultVehicle().getPosition()
					pRot = p.getDefaultVehicle().getRotation()

					#### Get all grabbed objects and move them where they need to be
					if p.grabbed:
						if p.grabbed[0].isValid():
							pCam = self.getSoldierCam(p)
							pCamRot = pCam.getRotation()
							sPos = self.getVectoredPosition(pPos,p.grabbed[1],pRot,pCamRot)
							if p.gridSnap < 3:
								sPos = (round(sPos[0],p.gridSnap),round(sPos[1],p.gridSnap),round(sPos[2],p.gridSnap))
							if p.grabbed[0].getPosition() != sPos:
								p.grabbed[0].setPosition(sPos)
								self.dbg("new position: " + str(sPos[0]) + "," + str(sPos[1]) + "," + str(sPos[2]))
							p.grabbed[0].setRotation(p.grabbed[4])
						else:
							p.grabbed = []

					if self.is_spGame and p.preGrabMode == 1 and self.timerCounter == 0:
						pCam = self.getSoldierCam(p)
						shortest = []
						shortest = [0,p.tracerDistance + 10.0]
						pCamRot = pCam.getRotation()
						sPos = self.getVectoredPosition(pPos,p.tracerDistance,pRot,pCamRot)
						sRot = pRot
						for v in p.objectTypes:
							if self.allowed_objects.count(v):
								for q in bf2.objectManager.getObjectsOfTemplate(v):
									if q.isValid():
										qDist = self.getVectorDistance(sPos,q.getPosition())
										if qDist < shortest[1]:
											shortest = [q,qDist]
						if p.preGrab:
							if shortest[1] > p.preGrab[1]:
								shortest[0] = p.preGrab[0]
								shortest[1] = p.preGrab[1]
						else:
							p.preGrab = [shortest[0],shortest[1],1,shortest[0].getPosition(),pCam.getRotation()]
						if p.preGrab[0] != shortest[0]:
							# Object, current flash thing, object position, player rotation
							p.preGrab[0].setPosition(p.preGrab[3])
							p.preGrab = [shortest[0],shortest[1],1,shortest[0].getPosition(),pCam.getRotation()]
						else:
							if p.preGrab[1] == 1:
								o = p.preGrab[0]
								if o.isValid():
									if hasattr(o,"oId"):
										host.rcon_invoke("Object.active id" + str(o.oId))
										host.rcon_invoke("Object.isVisible 0")
									else:
										o.setPosition((0.0,83453.0,0.0))
								p.preGrab[1] = 2
							else:
								o = p.preGrab[0]
								if o.isValid():
									if hasattr(o,"oId"):
										host.rcon_invoke("Object.active id" + str(o.oId))
										host.rcon_invoke("Object.isVisible 1")
									else:
										o.setPosition(p.preGrab[3])
								p.preGrab[1] = 1

			#### Refreeze all frozen objects
			invalidFreezes = []
			for o in self.obj_freeze:
				if self.obj_freeze[o]:
					if (o.isValid()) and (type(self.obj_freeze[o][0]) == type(('tuple','thing'))):
						if o.getPosition() != self.obj_freeze[o][0]:
							o.setPosition(self.obj_freeze[o][0])
					else:
						invalidFreezes.append(o)

			for f in invalidFreezes:
				if self.obj_freeze.has_key(f):
					self.obj_freeze.pop(f)

			#### Get number of players to denote spGame
			if bf2.playerManager.getNumberOfPlayers() < 2:
				#self.is_spGame = 1
				self.is_spGame = 1
			else:
				self.is_spGame = 0
		except:
			ExceptionOutput()

		try:
			if self.timerLoaderActive:
				try:
					self.timerLoaderObjects[self.timerLoader[1]]
				except:
					try:
						#self.obj_freeze.update(self.timerLoaderTempFreeze)
						self.setAllowGrab(True)
						self.setAllowCreate(True)
						self.timerLoaderActive = 0
						self.timerLoader = [1,0,""]
						self.timerLoaderObjects = []
						self.timerLoaderTempFreeze = {}
						#self.obj_freeze = None
						#self.obj_freeze = {}
						self.sendMsg(0,"Loading finished.")
						self.dbg("Load finalized.")
						return 0
					except:
						ExceptionOutput()
				if self.timerLoader[0] == 1:
					oName = self.timerLoaderObjects[self.timerLoader[1]][0]
					oPos = self.timerLoaderObjects[self.timerLoader[1]][1]
					oRot = self.timerLoaderObjects[self.timerLoader[1]][2]
					self.timerLoader[2] = oName
					self.dbg("Created spawn template - " + str(oName))
					self.createSpawnerObject(str(oName), oPos, oRot)
					self.timerLoader[0] = 2
					#self.sendMsg(0,"Step 1 of object " + str(self.timerLoader[1]))
				elif self.timerLoader[0] == 2:
					#if not extended_objects.count(timerLoaderObjects[timerLoader[1]][0]):
					proper_obj = 0
					for q in bf2.objectManager.getObjectsOfTemplate(self.timerLoaderObjects[self.timerLoader[1]][0]):
						if q.isValid():
							proper_obj = q
					oName = self.timerLoaderObjects[self.timerLoader[1]][0]
					oPos = self.timerLoaderObjects[self.timerLoader[1]][1]
					oRot = self.timerLoaderObjects[self.timerLoader[1]][2]
					if proper_obj and not self.ragdoll_objects.count(oName.lower()):
						#self.timerLoaderTempFreeze[proper_obj] = [(float(oPos[0]),float(oPos[1]),float(oPos[2])),(float(oRot[0]),float(oRot[1]),float(oRot[2]))]
						self.obj_freeze[proper_obj] = [(float(oPos[0]),float(oPos[1]),float(oPos[2])),(float(oRot[0]),float(oRot[1]),float(oRot[2]))]
					self.timerLoader[0] = 1
					self.timerLoader[1] += 1
					#self.sendMsg(0,"Step 2 of object " + str(self.timerLoader[1]))
		except:
			ExceptionOutput()

	def createSpawnerObject(self, sName, sPos, sRot = (0.0,0.0,0.0), user=0):
		if not self.spawnerTemplates.count(sName):
			host.rcon_invoke("ObjectTemplate.create ObjectSpawner sbx_" + sName)
			host.rcon_invoke("ObjectTemplate.activeSafe ObjectSpawner sbx_" + sName)
			host.rcon_invoke("ObjectTemplate.hasMobilePhysics 0")
			host.rcon_invoke("ObjectTemplate.setObjectTemplate 1 " + sName)
			host.rcon_invoke("ObjectTemplate.setObjectTemplate 2 " + sName)
			host.rcon_invoke("ObjectTemplate.minSpawnDelay 700000")
			host.rcon_invoke("ObjectTemplate.maxSpawnDelay 900000")
			self.spawnerTemplates.append(sName)
		host.rcon_invoke("Object.create sbx_" + sName)
		host.rcon_invoke("Object.absolutePosition " + str(sPos[0]) + "/" + str(sPos[1]) + "/" + str(sPos[2]))
		host.rcon_invoke("Object.rotation " + str(sRot[0]) + "/" + str(sRot[1]) + "/" + str(sRot[2]))
		host.rcon_invoke("Object.team 2")
		host.rcon_invoke("Object.delete")
		self.lastObject = [str(sName),user,sPos,sRot]

	def createSpawnerTempl(self, id, tmpl):
		host.rcon_invoke("ObjectTemplate.create ObjectSpawner sbx_" + id)
		host.rcon_invoke("ObjectTemplate.activeSafe ObjectSpawner sbx_" + id)
		host.rcon_invoke("ObjectTemplate.hasMobilePhysics 0")
		host.rcon_invoke("ObjectTemplate.setObjectTemplate 1 " + tmpl)
		host.rcon_invoke("ObjectTemplate.setObjectTemplate 2 " + tmpl)
		host.rcon_invoke("ObjectTemplate.minSpawnDelay 700000")
		host.rcon_invoke("ObjectTemplate.maxSpawnDelay 900000")

	def createSpawner(self, id, sPos, sRot):
		host.rcon_invoke("Object.create sbx_" + id)
		host.rcon_invoke("Object.absolutePosition " + str(sPos[0]) + "/" + str(sPos[1]) + "/" + str(sPos[2]))
		host.rcon_invoke("Object.rotation " + str(sRot[0]) + "/" + str(sRot[1]) + "/" + str(sRot[2]))
		host.rcon_invoke("Object.team 2")
		host.rcon_invoke("Object.delete")

	def onChatMessage(self, playerId, text, channel, flags):
		text = text.replace( "HUD_TEXT_CHAT_TEAM", "" )
		text = text.replace( "HUD_TEXT_CHAT_SQUAD", "" )
		text = text.replace( "HUD_CHAT_DEADPREFIX", "" )
		text = text.replace( "*\xA71DEAD\xA70*", "" )
		if text[0:1] == "!":
			pattern = re.compile(r'!(\w*) ?(.*)')
			matches = pattern.findall(text)
			command = matches[0][0]
			if matches[0][1] != "":
				parameter = matches[0][1]
			else:
				parameter = None
			if playerId == -1: #fix for local servers
				playerId = 255
			p = bf2.PlayerManager.Player(playerId)
			p = self.p_objects[p.getName()]
			#p = self.p_objects[playerId]
			if self.MP_LS_USERS.count(p.getName()) or self.is_spGame == 1:
				if command == "save" and parameter:
					self.saveSandbox(parameter)
				if command == "load" and parameter:
					if self.is_spGame:
						self.loadSandbox(parameter,p)
					else:
						self.advloadSandbox(parameter,playerId)
				if command == "clear":
					host.rcon_invoke("admin.restartmap")

	def saveSandbox(self, filename):
		file_string = str(host.sgl_getMapName()) + "," + str(self.SANDBOX_VERSION) + "@"
		out_file = open("mods/sandbox/saved/" + str(filename) + ".bf2sbx","w")
		increment = 0
		for t in self.object_types:
			for q in bf2.objectManager.getObjectsOfTemplate(t):
				if q.isValid():
					if not self.deleted_objects.count(q):
						increment += 1
						qPos = q.getPosition()
						qRot = q.getRotation()
						qNam = q.templateName
						file_string += str(qNam) + ";" + str(qPos[0]) + "," + str(qPos[1]) + "," + str(qPos[2]) + ";" + str(qRot[0]) + "," + str(qRot[1]) + "," + str(qRot[2]) + "!"
		out_file.write(file_string)
		out_file.close()
		self.sendMsg(0, "Your sandbox has been saved as " + str(filename) + ".bf2sbx")
		self.dbg("Saved file, " + str(increment) + " objects")

	def loadSandbox(self, filename, p):
		try: in_file = open("mods/sandbox/saved/" + str(filename) + ".bf2sbx","r")
		except:
			self.sendMsg(0, "The file you specified was not found.")
			return 0
		savedStr = in_file.read()
		a = savedStr.split("@")
		w = a[0].split(",")
		if not w.count(self.SANDBOX_VERSION):
			self.sendMsg(0, "This file was created with a different version of BF2:Sandbox. Some objects may be loaded incorrectly.")
		if w[0].lower() == str(host.sgl_getMapName()).lower():
			b = a[1].split("!")
			increment = 0
			for objInfo in b:
				if objInfo:
					objInfoList = objInfo.split(";")
					oName = objInfoList[0]
					oPos = objInfoList[1].split(",")
					oRot = objInfoList[2].split(",")
					if self.mp_objects.count(oName.lower()) and not self.is_spGame:
						oName += "_mp"
					#self.createObject(oName, oPos, oRot)
					self.createSpawnerObject(oName, oPos, oRot)
					if p.objectTypes.count(oName) < 1:
						p.objectTypes.append(oName)
					if self.object_types.count(oName) < 1:
						self.object_types.append(oName)
					increment += 1
			self.sendMsg(0, str(filename) + ".bf2sbx has been loaded.")
			self.dbg("Loaded file, " + str(increment) + " objects")
		else:
			self.sendMsg(0, "The file you specified was not created in this map.")
		in_file.close()

	def advloadSandbox(self, filename, playerId):
		p = bf2.PlayerManager.Player(playerId)
		p = self.p_objects[p.getName()]
		#p = self.p_objects[playerId]
		try: in_file = open("mods/sandbox/saved/" + str(filename) + ".bf2sbx","r")
		except:
			self.sendMsg(0,"The file you specified was not found.")
			return 0
		savedStr = in_file.read()
		a = savedStr.split("@")
		w = a[0].split(",")
		if not w.count(self.SANDBOX_VERSION):
			self.sendMsg(0, "This file was created with a different version of BF2:Sandbox. Some objects may be loaded incorrectly.")
		if not self.is_spGame:
			self.sendMsg(0, "Loading a sandbox into a multiplayer game may cause unpredictable behavior.")
		if w[0].lower() == str(host.sgl_getMapName()).lower():
			self.timerLoaderActive = 0
			self.timerLoader = [1,0,""]
			self.timerLoaderObjects = []
			b = a[1].split("!")
			self.dbg("Starting object loops")
			increment = 0
			for objInfo in b:
				if objInfo:
					self.dbg("Found an object")
					objInfoList = objInfo.split(";")
					oName = objInfoList[0]
					oPos = objInfoList[1].split(",")
					oRot = objInfoList[2].split(",")
					oPos = (float(oPos[0]),float(oPos[1]),float(oPos[2]))
					oRot = (float(oRot[0]),float(oRot[1]),float(oRot[2]))
					if self.mp_objects.count(oName.lower()) and not self.is_spGame:
						oName += "_mp"
					self.timerLoaderObjects.append([oName,oPos,oRot])
					if p.objectTypes.count(oName) < 1:
						p.objectTypes.append(oName)
					if self.object_types.count(oName) < 1:
						self.object_types.append(oName)
					increment += 1
			self.sendMsg(0,str(filename) + ".bf2sbx loading using experimental MP mode")
			self.setAllowGrab(False)
			self.setAllowCreate(False)
			self.timerLoaderActive = 1
		else:
			self.sendMsg(0,"The file you specified was not created in this map.")
		in_file.close()

	def fixPlayerObject(self, pId, pName):
		try:
			old = self.p_objects[pName]
			new = bf2.PlayerManager.Player(pId)

			new.tracerDistance = old.tracerDistance
			new.objects = old.objects
			new.objectTypes = old.objectTypes
			new.grabbed = old.grabbed
			new.lastRotation = old.lastRotation
			new.lastSpawned = old.lastSpawned
			new.lastTemplate = old.lastTemplate
			new.tracerOn = old.tracerOn
			new.tracerDistance = old.tracerDistance
			new.gridSnap = old.gridSnap
			new.rotSnap = old.rotSnap
			new.autoGrab = old.autoGrab
			new.numSpawned = old.numSpawned
			new.preGrab = old.preGrab
			new.preGrabMode = old.preGrabMode
			new.lastAction = old.lastAction
			new.spawnDelay = old.spawnDelay

			self.p_objects[pName] = new
		except:
			ExceptionOutput()

	def onEnterVehicle(self, player, vehicle, freeSoldier = False):
		p = self.p_objects[player.getName()]
		p.grabbed = []

	def onExitVehicle(self, player, vehicle):
		p = self.p_objects[player.getName()]
		p.grabbed = []

	def onPlayerDeathCQ(self, victim, vehicle):
		p = self.p_objects[victim.getName()]
		p.grabbed = []

	def onPlayerKilledCQ(self, victim, attacker, weapon, assists, object):
		p = self.p_objects[victim.getName()]
		p.grabbed = []

	def onPlayerDisconnect(self, player):
		if self.p_objects.has_key(player.getName()):
			pl = p_objects[player.getName()]
			pl.grabbed = []
			for o in pl.objects:
				host.rcon_invoke("Object.active id" + str(o.oId))
				host.rcon_invoke("Object.delete")
		for o in self.obj_locked:
			if self.obj_locked[o] == player.index:
				self.obj_locked.pop(o)
		try:
			self.p_objects.pop(player.getName())
		except:
			ExceptionOutput()

	def getVectoredPosition(self, pPos, distance, pRot, pCamRot):
		from math import pi
		#Calculate the height using a triangle
		p = pCamRot[1] #Pitch of the player's cam
		a = 90.0 - abs(p)
		p = pi * p / 180.0
		a = pi * a / 180.0
		sVz = (distance * math.sin(p))/math.sin(a)
		#convert from crazy-ass BF2 values to polar coordinate values
		new_pRot = -(pRot[0] + pCamRot[0]) + 90.0
		sVx,sVy = self.rect(distance,new_pRot,1) #Convert vector of [spawnDistance] with heading of cam rotation to x,y values
		sPos = (pPos[0] + sVx, pPos[1] - sVz, pPos[2] + sVy)
		return sPos

	def onPlayerSpawn(self, player, soldier):
		#if not self.p_objects.has_key(player.index):
		#self.p_objects[player.index] = player
		#self.p_objects[player.getName()] = bf2.PlayerManager.Player(player.index)
		#self.dbg("Set a pobject for " + str(player.getName()))
		pass #Do nothing

	def rect(self, r, w, deg=0): # radian if deg=0; degree if deg=1
		from math import cos, sin, pi
		if deg:
			w = pi * w / 180.0
		return r * cos(w), r * sin(w)

	def sendMsg(self, p, msg):
		if p:
			host.rcon_invoke("game.sayAll \"" + str(p.getName()) + " - " + str(msg) + "\"")
		else:
			host.rcon_invoke("game.sayAll \"" + str(msg) + "\"")

	def getVectorDistance(self, pos1, pos2):
		diffVec = [0.0, 0.0, 0.0]
		diffVec[0] = math.fabs(pos1[0] - pos2[0])
		diffVec[1] = math.fabs(pos1[1] - pos2[1])
		diffVec[2] = math.fabs(pos1[2] - pos2[2])
		return math.sqrt(diffVec[0] * diffVec[0] + diffVec[1] * diffVec[1] + diffVec[2] * diffVec[2])

	def getSoldierCam(self, player):
		p = getRootParent(player.getDefaultVehicle())
		kids = p.getChildren()
		for child in kids:
			if child.templateName == "SoldierCamera":
				return child

	def disableBoundaries(self):
		host.rcon_invoke("CombatAreaManager.use 0")
		host.rcon_invoke("CombatAreaManager.timeAllowedOutside 1000.000000")
		self.sBoundaries = False

	def enableBoundaries(self):
		host.rcon_invoke("CombatAreaManager.use 1")
		host.rcon_invoke("CombatAreaManager.timeAllowedOutside 10.00")
		self.sBoundaries = True

	def setAllowedObjects(self,objList):
		if type(self.allowed_objects) == type(objList):
			self.allowed_objects = objList

	def addAllowedObjects(self,obj):
		# This can take either a list or a string.
		if type(self.allowed_objects) == type(obj):
			self.allowed_objects.extend(obj)
		else:
			self.allowed_objects.append(obj)

	def delAllowedObjects(self,obj):
		# This can take either a list or a string.
		if type(self.allowed_objects) == type(obj):
			for x in obj:
				try:
					self.allowed_objects.remove(x)
				except:
					pass #do nothing
		else:
			self.allowed_objects.remove(obj)

	def setSpawnLimits(self,objSp,ragdSp):
		self.SPAWN_LIMIT = objSp
		self.RAGDOLL_LIMIT = ragdSp

	def setMpAdmins(self,adminList):
		if type(self.allowed_objects) == type(adminList):
			self.MP_LS_USERS = adminList
		else:
			self.MP_LS_USERS = [adminList]

	def setAllowLock(self,val):
		if val:
			self.allow_lock = True
		else:
			self.allow_lock = False

	def setAllowGrab(self,val):
		if val:
			self.allow_grab = True
		else:
			self.allow_grab = False

	def setAllowCreate(self,val,override=True):
		if val:
			self.allow_create = True
		else:
			self.allow_create = False
		if override:
			self.allowCreateOverride = True
		else:
			self.allowCreateOverride = False

	def setMinPlayers(self,val):
		host.rcon_invoke("sv.numPlayersNeededToStart " + str(val))

	def createObject(self, oName, oPos, oRot = (0.0,0.0,0.0), user = 0):
		try:
			oPos = (float(oPos[0]),float(oPos[1]),float(oPos[2]))
			oRot = (float(oRot[0]),float(oRot[1]),float(oRot[2]))
			host.rcon_invoke("gameLogic.createObject " + str(oName))
			self.lastObject = [str(oName),user,oPos,oRot]
			#for q in bf2.objectManager.getObjectsOfTemplate(str(oName)):
			#	if q.isValid():
			#		lastObj = q
			#objs = host.rcon_invoke("object.listObjectsofTemplate " + str(oName))
			#objs = objs.split()
			#lastObj.oId = objs[-7]
			#lastObj.setPosition(oPos)
			#lastObj.setRotation(oRot)
			#return lastObj
			return 1
		except:
			ExceptionOutput()

	def deleteObject(self, o):
		try:
			if o:
				if o.isValid():
					if hasattr(o,"oId"):
						host.rcon_invoke("Object.active id" + str(o.oId))
						host.rcon_invoke("Object.delete")
					else:
						o.setPosition((0.0,99999.0,0.0))
						self.deleted_objects.append(o)
		except:
			ExceptionOutput()

def start():
	return sandboxCore()

print "Sandbox core loaded."
